# Copyright (c) 2025 The OSCAL Compass Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""cis-xlsx-to-oscal-cd task tests."""

import configparser
import os
import pathlib
from typing import Dict
from unittest.mock import patch

from tests.test_utils import TEST_DIR, set_cwd_unsafe

import trestle.tasks.cis_xlsx_to_oscal_cd as cis_xlsx_to_oscal_cd
from trestle.oscal.component import ComponentDefinition
from trestle.tasks.base_task import TaskOutcome

root_dir = TEST_DIR / '../'
db2_config = TEST_DIR / 'data/tasks/cis-xlsx-to-oscal-cd/test-cis-xlsx-to-oscal-cd.db2.snippet.config'


def _get_section(tmp_path: pathlib.Path, file_: str) -> Dict:
    """Get section."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(file_)
    config.read(config_path)
    section = config['task.cis-xlsx-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    return section


def test_cis_xlsx_to_oscal_cd_compare(tmp_path: pathlib.Path):
    """Test compare."""
    x = cis_xlsx_to_oscal_cd.SortHelper.compare('A', 'B')
    assert x == -1
    x = cis_xlsx_to_oscal_cd.SortHelper.compare('0.0', '1.0')
    assert x == -1
    x = cis_xlsx_to_oscal_cd.SortHelper.compare('1.0', '0.0')
    assert x == 1
    x = cis_xlsx_to_oscal_cd.SortHelper.compare('1.1', '1.1')
    assert x == 0


def test_cis_xlsx_to_oscal_cd_print_info(tmp_path: pathlib.Path):
    """Test print_info call."""
    section = _get_section(tmp_path, db2_config)
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.print_info()
    assert retval is None


def test_cis_xlsx_to_oscal_cd_simulate(tmp_path: pathlib.Path):
    """Test simulate call."""
    section = _get_section(tmp_path, db2_config)
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute(tmp_path: pathlib.Path):
    """Test execute call - db2."""
    section = _get_section(tmp_path, db2_config)
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_db2(tmp_path)


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_combined(tmp_path: pathlib.Path):
    """Test execute call - db2."""
    section = _get_section(tmp_path, db2_config)
    section['benchmark-file'
            ] = 'tests/data/tasks/cis-xlsx-to-oscal-cd/CIS_IBM_Db2_11_Benchmark_v1.1.0.snippet_combined.xlsx'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_db2(tmp_path)


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_missing_column(tmp_path: pathlib.Path):
    """Test execute call - missing column."""
    section = _get_section(tmp_path, db2_config)
    section['benchmark-file'
            ] = 'tests/data/tasks/cis-xlsx-to-oscal-cd/CIS_IBM_Db2_11_Benchmark_v1.1.0.snippet_missing_column.xlsx'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_bad_config(tmp_path: pathlib.Path):
    """Test execute call - bad config."""
    section = _get_section(tmp_path, db2_config)
    del section['benchmark-file']
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_bad_overwrite(tmp_path: pathlib.Path):
    """Test execute call - bad overwrite."""
    section = _get_section(tmp_path, db2_config)
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    section['output-overwrite'] = 'false'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_merge(tmp_path: pathlib.Path):
    """Test execute call - merge."""
    section = _get_section(tmp_path, db2_config)
    section['benchmark-file'
            ] = 'tests/data/tasks/cis-xlsx-to-oscal-cd/CIS_IBM_Db2_11_Benchmark_v1.1.0.snippet_merge.xlsx'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_db2(tmp_path)


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_rule_prefix(tmp_path: pathlib.Path):
    """Test execute call - rule prefix."""
    section = _get_section(tmp_path, db2_config)
    section['benchmark-rule-prefix'] = 'CIS'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_db2(tmp_path)


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_control_prefix(tmp_path: pathlib.Path):
    """Test execute call - control prefix."""
    section = _get_section(tmp_path, db2_config)
    section['benchmark-control-prefix'] = 'cisc'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_db2(tmp_path)


def test_cis_xlsx_to_oscal_cd_execute_control_bad(tmp_path: pathlib.Path):
    """Test execute call - control bad."""
    section = _get_section(tmp_path, db2_config)
    section['benchmark-file'
            ] = 'tests/data/tasks/cis-xlsx-to-oscal-cd/CIS_IBM_Db2_11_Benchmark_v1.1.0.snippet_bad_control.xlsx'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


@set_cwd_unsafe(root_dir)
def test_cis_xlsx_to_oscal_cd_execute_columns_exclude(tmp_path: pathlib.Path):
    """Test execute call - control prefix."""
    section = _get_section(tmp_path, db2_config)
    section['columns-exclude'] = '"Recommendation #", "Profile", "Description"'
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    _validate_db2(tmp_path)


def test_cis_xlsx_to_oscal_cd_execute_config_missing(tmp_path: pathlib.Path):
    """Test execute call - config_missing."""
    section = None
    tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_cd_execute_count_mismatch(tmp_path: pathlib.Path):
    """Test execute call - count mismatch."""
    with patch('trestle.tasks.cis_xlsx_to_oscal_cd.CombineHelper._populate_combined_map') as mock_original_function:
        mock_original_function.return_value = -5
        section = _get_section(tmp_path, db2_config)
        tgt = cis_xlsx_to_oscal_cd.CisXlsxToOscalCd(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_cis_xlsx_to_oscal_cd_execute_csv_row_mgr(tmp_path: pathlib.Path):
    """Test execute call - csv row mgr."""
    row_names = ['row1', 'row2', 'row3']
    # create new row mgr
    csv_row_mgr = cis_xlsx_to_oscal_cd.CsvRowMgr(row_names)
    # test valid case
    try:
        csv_row_mgr.put('row1', '')
    except RuntimeError:
        assert 0 == 1
    # test invalid case
    try:
        csv_row_mgr.put('rowX', '')
        assert 0 == 1
    except RuntimeError:
        pass


def _validate_db2(tmp_path: pathlib.Path):
    """Validate produced OSCAL for db2 cd."""
    # read catalog
    file_path = tmp_path / 'component-definition.json'
    component_definition = ComponentDefinition.oscal_read(file_path)
    # spot check
    assert len(component_definition.components) == 1
    assert component_definition.metadata.title == 'CIS IBM Db2 11 Benchmark'
    assert component_definition.metadata.version == '1.1.0'
    component = component_definition.components[0]
    assert component.type == 'software'
    assert len(component.props) == 552
    prop = component.props[0]
    assert prop.name == 'Rule_Id'
    assert prop.ns == 'https://oscal-compass/compliance-trestle/schemas/oscal/cd'
    assert prop.value == 'CIS-1.1.1'
    assert prop.remarks == 'rule_set_00'
    assert len(component.control_implementations) == 1
    prop = component.props[551]
    assert prop.name == 'Group_Description_Level_1'
    assert prop.ns == 'https://oscal-compass/compliance-trestle/schemas/oscal/cd'
    assert prop.value.startswith('This section provides guidance on various database configuration parameters.')
    assert prop.remarks == 'rule_set_22'
    assert len(component.control_implementations) == 1
    control_implementation = component.control_implementations[0]
    assert len(control_implementation.implemented_requirements) == 6
    assert control_implementation.source == 'catalogs/CIS_controls_v8/catalog.json'
    assert control_implementation.description == 'CIS catalog v8'
    implemented_requirement = control_implementation.implemented_requirements[0]
    assert implemented_requirement.control_id == 'cisc-7.4'
    assert len(implemented_requirement.props) == 1
    prop = implemented_requirement.props[0]
    assert prop.name == 'Rule_Id'
    assert prop.ns == 'https://oscal-compass/compliance-trestle/schemas/oscal/cd'
    assert prop.value == 'CIS-1.1.1'
    implemented_requirement = control_implementation.implemented_requirements[5]
    assert implemented_requirement.control_id == 'cisc-3.10'



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2024 The OSCAL Compass Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle core.jinja extension loading functionality."""
import functools
import logging
from typing import List

from trestle.core.jinja import filters, tags
from trestle.core.jinja.base import Extension, TrestleJinjaExtension
from trestle.core.plugins import discovered_plugins

logger = logging.getLogger(__name__)


@functools.cache
def extensions() -> List[Extension]:
    """Return list of Jinja extensions packaged with compliance-trestle and included from plugins."""
    extensions = [tags.MDSectionInclude, tags.MDCleanInclude, tags.MDDatestamp, filters.JinjaSSPFilters]
    # This block is uncovered as trestle cannot find plugins in it's unit tests - it is the base module.
    for plugin, ext_cls in discovered_plugins('jinja_ext'):  # pragma: nocover
        # add extensions (derived from TrestleJinjaExtension) to extensions list
        if issubclass(ext_cls, Extension):
            # don't add Extension or TrestleJinjaExtension
            if ext_cls is not TrestleJinjaExtension and ext_cls is not Extension:
                extensions.append(ext_cls)
                logger.info(f'{ext_cls} added to jinja extensions from plugin {plugin}')
    return extensions



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2024 The OSCAL Compass Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin discovery code."""

import importlib
import inspect
import logging
import pathlib
import pkgutil
from typing import Any, Iterator, Tuple

logger = logging.getLogger(__name__)

_discovered_plugins = {
    name: importlib.import_module(name)
    for finder,
    name,
    ispkg in pkgutil.iter_modules()
    if name.startswith('trestle_')
}


def discovered_plugins(search_module: str) -> Iterator[Tuple[str, Any]]:
    """Yield discovered plugin classes within a given module name."""
    logger.debug(_discovered_plugins)
    # This block is uncovered as trestle cannot find plugins in it's unit tests - it is the base module.
    for plugin, value in _discovered_plugins.items():  # pragma: nocover
        for _, module, _ in pkgutil.iter_modules([pathlib.Path(value.__path__[0], search_module)]):
            logger.debug(module)
            plugin_module = importlib.import_module(f'{plugin}.{search_module}.{module}')
            clsmembers = inspect.getmembers(plugin_module, inspect.isclass)
            logger.debug(clsmembers)
            for _, plugin_cls in clsmembers:
                yield (plugin, plugin_cls)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2024 The OSCAL Compass Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle utilities to customize jinja filters."""
import logging
from typing import Any, Iterator, List, Optional

from jinja2.environment import Environment

from trestle.common.list_utils import as_list, get_default
from trestle.core.jinja.base import TrestleJinjaExtension
from trestle.oscal.common import Party, ResponsibleParty
from trestle.oscal.ssp import Diagram, SystemSecurityPlan

logger = logging.getLogger(__name__)


def first_or_none(value: Optional[List[Any]]) -> Optional[Any]:
    """Retrieve the first array entry, or None for lists that are None or empty."""
    return next(iter(as_list(value)), None)


def get_party(uuid: str, ssp: SystemSecurityPlan) -> Optional[Party]:
    """Get the metadata.parties entry for this UUID."""
    return next((x for x in as_list(ssp.metadata.parties) if x.uuid == uuid), None)


def parties_for_role(responsible_parties: List[ResponsibleParty], role_id: str,
                     ssp: SystemSecurityPlan) -> Iterator[Party]:
    """Get a list of parties from a list of responsible_parties and a given role_id."""
    logger.debug(f'Finding parties for role: {role_id}')
    for responsible_party in as_list(responsible_parties):
        if responsible_party.role_id == role_id:
            logger.debug(
                f'Found responsible party for role_id: {role_id} with {len(responsible_party.party_uuids)} parties'
            )
            for uuid in responsible_party.party_uuids:
                logger.debug(f'Looking for parties with uuid: {uuid}')
                party = get_party(uuid, ssp)
                if party:
                    yield party


def diagram_href(diagram: Optional[Diagram]) -> str:
    """Retrieve the diagrams's link href."""
    if diagram:
        return next((link.href for link in as_list(diagram.links) if link.rel == 'diagram'), '')
    else:
        return ''


class JinjaSSPFilters(TrestleJinjaExtension):
    """Collection of useful OSCAL-specific filters."""

    def __init__(self, environment: Environment) -> None:
        """Initialize class and add filters."""
        super(JinjaSSPFilters, self).__init__(environment)

        environment.filters['as_list'] = as_list
        environment.filters['get_default'] = get_default
        environment.filters['first_or_none'] = first_or_none
        environment.filters['get_party'] = get_party
        environment.filters['parties_for_role'] = parties_for_role
        environment.filters['diagram_href'] = diagram_href



# Copyright (c) 2025 The OSCAL Compass Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL transformation tasks."""
import configparser
import csv
import datetime
import logging
import pathlib
import tempfile
import traceback
from configparser import SectionProxy
from functools import cmp_to_key
from typing import Dict, Iterator, List, Optional

from openpyxl import load_workbook
from openpyxl.workbook.workbook import Workbook

from trestle.common.str_utils import as_bool
from trestle.tasks.base_task import TaskBase
from trestle.tasks.base_task import TaskOutcome
from trestle.tasks.csv_to_oscal_cd import CsvToOscalComponentDefinition

logger = logging.getLogger(__name__)

timestamp = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0).isoformat()

default_benchmark_control_prefix = 'cisc-'
default_benchmark_rule_prefix = 'CIS-'
output_file = 'component-definition.json'

head_recommendation_no = 'Recommendation #'
head_section_no = 'Section #'


class CisXlsxToOscalCd(TaskBase):
    """
    Task to transform CIS .xlsx to OSCAL component definition.

    Attributes:
        name: Name of the task.
    """

    name = 'cis-xlsx-to-oscal-cd'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task.

        Args:
            config_object: Config section associated with the task.
        """
        self.config_object = config_object
        super().__init__(config_object)
        self._default_benchmark_sheet_name = 'Combined Profiles'
        self._default_component_type = 'software'
        self._default_output_overwrite = 'True'
        #
        self._example_benchmark_file = 'data/CIS_IBM_Db2_11_Benchmark_v1.1.0.xlsx'
        self._example_benchmark_title = 'CIS IBM Db2 11 Benchmark'
        self._example_benchmark_version = '1.1.0'
        self._example_oscal_cd_dir = 'data/component-definitions/CIS_IBM_Db2_11_Benchmark_v1.1.0'
        self._example_namespace = 'https://oscal-compass/compliance-trestle/schemas/oscal/cd'
        self._example_profile_version = 'v8'
        self._example_profile_source = 'catalogs/CIS_controls_v8/catalog.json'
        self._example_profile_description = 'CIS catalog v8'
        self._example_component_name = 'Db2 11'

    def print_info(self) -> None:
        """Print the help string."""
        logger.info(f'Help information for {self.name} task.')
        logger.info('')
        logger.info('Purpose: Create component definition from standard CIS benchmark')
        logger.info('')
        logger.info('Configuration flags sit under [task.cis-xlsx-to-oscal-cd]:')
        text1 = '  benchmark-file             = '
        text2 = f'(required) path of file to read the CIS benchmark .xlsx, e.g., "{self._example_benchmark_file}".'
        logger.info(text1 + text2)
        text1 = '  benchmark-title            = '
        text2 = f'(required) title of the CIS benchmark, e.g., "{self._example_benchmark_title}".'
        logger.info(text1 + text2)
        text1 = '  benchmark-version          = '
        text2 = f'(required) version of the CIS benchmark .xlsx, e.g., "{self._example_benchmark_version}".'
        logger.info(text1 + text2)
        text1 = '  benchmark-control-prefix   = '
        text2 = f'(optional) benchmark control prefix, default = "{default_benchmark_control_prefix}".'
        logger.info(text1 + text2)
        text1 = '  benchmark-rule-prefix      = '
        text2 = f'(optional) benchmark rule prefix, default = "{default_benchmark_rule_prefix}".'
        logger.info(text1 + text2)
        text1 = '  benchmark-sheet-name       = '
        text2 = f'(optional) benchmark sheet name, default = "{self._default_benchmark_sheet_name}".'
        logger.info(text1 + text2)
        text1 = '  component-name             = '
        text2 = f'(required) component name, e.g., "{self._example_component_name}".'
        logger.info(text1 + text2)
        text1 = '  component-description      = '
        text2 = f'(required) component description, e.g., "{self._example_component_name}".'
        logger.info(text1 + text2)
        text1 = '  component-type             = '
        text2 = f'(required) component type, e.g., "{self._default_component_type}".'
        logger.info(text1 + text2)
        text1 = '  namespace                  = '
        text2 = f'(required) namespace, e.g., "{self._example_namespace}".'
        logger.info(text1 + text2)
        text1 = '  output-dir                 = '
        text2 = f'(required) path of folder to write the OSCAL {output_file}, e.g., "{self._example_oscal_cd_dir}".'
        logger.info(text1 + text2)
        text1 = '  output-overwrite           = '
        text2 = f'(optional) output overwrite, default = "{self._default_output_overwrite}".'
        logger.info(text1 + text2)
        text1 = '  profile-version            = '
        text2 = f'(required) profile version, e.g., "{self._example_profile_version}".'
        logger.info(text1 + text2)
        text1 = '  profile-source             = '
        text2 = f'(required) profile source, e.g., "{self._example_profile_source}".'
        logger.info(text1 + text2)
        text1 = '  profile-description        = '
        text2 = f'(required) profile description, e.g., "{self._example_profile_description}".'
        logger.info(text1 + text2)

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        return TaskOutcome('simulated-success')

    def execute(self) -> TaskOutcome:
        """Provide an actual outcome."""
        try:
            return self._execute()
        except Exception:
            logger.info(traceback.format_exc())
            return TaskOutcome('failure')

    def _execute(self) -> TaskOutcome:
        """Wrap the execute for exception handling."""
        if not self.config_object:
            logger.warning('config missing')
            return TaskOutcome('failure')
        # required
        try:
            self._benchmark_file = self.config_object['benchmark-file']
            self._oscal_cd_dir = self.config_object['output-dir']
            self._namespace = self.config_object['namespace']
            self._profile_version = self.config_object['profile-version']
        except KeyError as e:
            logger.info(f'key {e.args[0]} missing')
            return TaskOutcome('failure')
        # output
        self._oscal_cd_path = pathlib.Path(self._oscal_cd_dir)
        # insure output dir exists
        self._oscal_cd_path.mkdir(exist_ok=True, parents=True)
        # calculate output file name & check writability
        oname = 'component-definition.json'
        ofile = self._oscal_cd_path / oname
        overwrite = self.config_object.get('output-overwrite', self._default_output_overwrite)
        overwrite = as_bool(overwrite)
        if not overwrite and pathlib.Path(ofile).exists():
            logger.warning(f'output: {ofile} already exists')
            return TaskOutcome('failure')
        with self._get_tempdir() as tmpdir:
            # step 1 - add combined sheet, if needed
            combine_helper = CombineHelper(self.config_object, tmpdir)
            combine_helper.run()
            # step 2 - create trestle ready csv file from xlsx file
            xlsx_to_csv_helper = XlsxToCsvHelper(self.config_object, tmpdir)
            xlsx_to_csv_helper.run()
            # step 3 - create OSCAL json file from csv file
            csv_to_json_helper = CsvToJsonHelper(self.config_object, tmpdir)
            task_outcome = csv_to_json_helper.run()
            return task_outcome

    def _get_tempdir(self) -> tempfile.TemporaryDirectory():
        """Get tmpdir."""
        return tempfile.TemporaryDirectory()


class SortHelper:
    """SortHelper."""

    @staticmethod
    def compare(item1: str, item2: str) -> int:
        """Compare."""
        # get parts
        parts1 = ''.split('.')
        if item1 is not None:
            parts1 = str(item1).split('.')
        parts2 = ''.split('.')
        if item2 is not None:
            parts2 = str(item2).split('.')
        # normalize parts length
        while len(parts1) < len(parts2):
            parts1.append('0')
        while len(parts2) < len(parts1):
            parts2.append('0')
        # comparison
        rval = 0
        for i in range(len(parts1)):
            try:
                v1 = int(parts1[i])
            except Exception:
                rval = -1
                break
            try:
                v2 = int(parts2[i])
            except Exception:
                rval = 1
                break
            if v1 < v2:
                rval = -1
                break
            if v1 > v2:
                rval = 1
                break
        text = f'compare rval: {rval} item1: {item1} item2: {item2}'
        logger.debug(f'{text}')
        return rval


class SheetHelper:
    """SheetHelper."""

    def __init__(self, wb: Workbook, sn: str) -> None:
        """Initialize."""
        self.wb = wb
        self.sn = sn
        self.ws = self.wb[self.sn]

    def get_sn(self) -> int:
        """Get sheet name."""
        return self.sn

    def get_max_col(self) -> int:
        """Get max column."""
        return self.ws.max_column

    def row_generator(self) -> Iterator[int]:
        """Generate rows until max reached."""
        row = 1
        while row <= self.ws.max_row:
            yield row
            row += 1

    def get_cell_value(self, row: int, col: int) -> str:
        """Get cell value for given row and column name."""
        cell = self.ws.cell(row, col)
        return cell.value

    def put_cell_value(self, row: int, col: int, value: str) -> None:
        """Get cell value for given row and column name."""
        cell = self.ws.cell(row, col)
        cell.value = value

    @staticmethod
    def get_sheetname_prefixes() -> List[str]:
        """Get sheetnames prefixes."""
        rval = ['Level 1', 'Level 2']
        return rval

    @staticmethod
    def get_sheetname() -> str:
        """Get sheetname output."""
        rval = 'Combined Profiles'
        return rval


class ColHelper:
    """Col Helper."""

    @staticmethod
    def get_section() -> int:
        """Get section col no."""
        return 1

    @staticmethod
    def get_recommendation() -> int:
        """Get recommendation col no."""
        return 2


class Int:
    """Int."""

    def __init__(self, value=0):
        """Initialize."""
        self.value = value

    def inc_value(self) -> None:
        """Increment."""
        self.value += 1

    def get_value(self) -> int:
        """Get."""
        return self.value


class CombineHelper:
    """Combine helper."""

    tgt_col_profile = 3

    def __init__(self, config: SectionProxy, tmpdir: str) -> None:
        """Initialize."""
        benchmark_file = config['benchmark-file']
        self.ipath = pathlib.Path(benchmark_file)
        self.opath = pathlib.Path(tmpdir) / self.ipath.name
        self.wb = load_workbook(self.ipath)
        self.ws_map = {}
        self.combined_map = {}

    def run(self) -> None:
        """Run."""
        self._add_sheet_combined_profiles()
        self._save()

    def _gather_sheets(self) -> None:
        """Gather sheets."""
        for sn in self.wb.sheetnames:
            for pn in self.sheetnames_prefixes:
                if sn.startswith(pn):
                    self.ws_map[sn] = SheetHelper(self.wb, sn)
                    logger.debug(f'input sheet {sn} to be combined.')
                    break

    def _validate_columns_count(self) -> None:
        """Validate columns count."""
        columns = -1
        for sn in self.ws_map.keys():
            sheet_helper = self.ws_map[sn]
            if columns < 0:
                columns = sheet_helper.get_max_col()
            if columns != sheet_helper.get_max_col():
                raise RuntimeError(f'{sn} unexpected columns count {sheet_helper.get_max_col()} for sheet {sn}')

    def _populate_combined_map(self) -> int:
        """Populate combined map."""
        src_col_section_no = ColHelper.get_section()
        src_col_recommendation_no = ColHelper.get_recommendation()
        rec_count_sheets = 0
        # populate combined map
        for sn in self.ws_map.keys():
            sheet_helper = SheetHelper(self.wb, sn)
            # process all rows from individual sheet
            rec_count_sheets += self._process_sheet(sheet_helper, src_col_section_no, src_col_recommendation_no)
        return rec_count_sheets

    def _process_sheet(self, sheet_helper: SheetHelper, src_col_section_no: int, src_col_recommendation_no: int) -> int:
        """Process sheet."""
        rec_count = 0
        for row in sheet_helper.row_generator():
            # section
            section_no = sheet_helper.get_cell_value(row, src_col_section_no)
            if section_no not in self.combined_map.keys():
                self.combined_map[section_no] = {}
            # recommendation
            recommendation_no = sheet_helper.get_cell_value(row, src_col_recommendation_no)
            if recommendation_no not in self.combined_map[section_no].keys():
                self.combined_map[section_no][recommendation_no] = {}
            # combine head or data
            if row == 1:
                self._combine_head(sheet_helper, row, section_no, recommendation_no, CombineHelper.tgt_col_profile)
            else:
                self._combine_data(sheet_helper, row, section_no, recommendation_no, CombineHelper.tgt_col_profile)
                if recommendation_no:
                    rec_count += 1
        return rec_count

    def _handle_head_row(
        self, combined_helper: SheetHelper, row: Int, kvset: Dict, section_no: str, recommendation_no: str
    ) -> None:
        """Handle head row."""
        for col in kvset.keys():
            value = self.combined_map[section_no][recommendation_no][col]
            if col == CombineHelper.tgt_col_profile:
                value = value[0]
            combined_helper.put_cell_value(row.get_value(), col, value)
        row.inc_value()

    def _handle_data_row_control(
        self,
        combined_helper: SheetHelper,
        row: Int,
        kvset: Dict,
        section_no: str,
        recommendation_no: str,
        rec_count_merged: Int
    ) -> None:
        """Handle data row control."""
        profiles = kvset[CombineHelper.tgt_col_profile]
        for profile in profiles:
            for col in kvset.keys():
                value = self.combined_map[section_no][recommendation_no][col]
                if col == CombineHelper.tgt_col_profile:
                    value = profile
                combined_helper.put_cell_value(row.get_value(), col, value)
            row.inc_value()
            rec_count_merged.inc_value()

    def _handle_data_row_non_control(
        self, combined_helper: SheetHelper, row: Int, kvset: Dict, section_no: str, recommendation_no: str
    ) -> None:
        """Handle data row non-control."""
        for col in kvset.keys():
            value = self.combined_map[section_no][recommendation_no][col]
            if col == CombineHelper.tgt_col_profile:
                value = None
            combined_helper.put_cell_value(row.get_value(), col, value)
        row.inc_value()

    def _handle_data_row(
        self,
        combined_helper: SheetHelper,
        row: Int,
        kvset: Dict,
        section_no: str,
        recommendation_no: str,
        rec_count_merged: Int
    ) -> None:
        """Handle data row."""
        if recommendation_no:
            self._handle_data_row_control(combined_helper, row, kvset, section_no, recommendation_no, rec_count_merged)
        else:
            self._handle_data_row_non_control(combined_helper, row, kvset, section_no, recommendation_no)

    def _populate_combined_sheet(self, combined_helper: SheetHelper) -> int:
        """Populate combined sheet."""
        rec_count_merged = Int(0)
        row = Int(1)
        keys1 = list(self.combined_map.keys())
        keys1.sort(key=cmp_to_key(SortHelper.compare))
        for section_no in keys1:
            section = self.combined_map[section_no]
            keys2 = list(section.keys())
            keys2.sort(key=cmp_to_key(SortHelper.compare))
            for recommendation_no in keys2:
                kvset = self.combined_map[section_no][recommendation_no]
                if row.get_value() == 1:
                    self._handle_head_row(combined_helper, row, kvset, section_no, recommendation_no)
                else:
                    self._handle_data_row(combined_helper, row, kvset, section_no, recommendation_no, rec_count_merged)
        return rec_count_merged.get_value()

    def _add_sheet_combined_profiles(self) -> None:
        """Add sheet combined profiles."""
        # output sheet
        self.sheetname_output = SheetHelper.get_sheetname()
        exists = self.sheetname_output in self.wb.sheetnames
        if exists:
            logger.debug(f'output sheet {self.sheetname_output} exists.')
            return
        # input sheets
        self.sheetnames_prefixes = SheetHelper.get_sheetname_prefixes()
        # sheets
        self._gather_sheets()
        # validate
        self._validate_columns_count()
        # key columns mappings
        rec_count_sheets = self._populate_combined_map()
        # add combined sheet
        sn = self.sheetname_output
        self.wb.create_sheet(sn)
        combined_helper = SheetHelper(self.wb, sn)
        self.ws_map[sn] = combined_helper
        # populate combined sheet
        rec_count_merged = self._populate_combined_sheet(combined_helper)
        # correctness check
        if rec_count_sheets != rec_count_merged:
            raise RuntimeError(f'recommendation counts original: {rec_count_sheets} merged: {rec_count_merged}')

    def _combine_head(
        self, sheet_helper: SheetHelper, row: int, section_no: str, recommendation_no: str, tgt_col_profile: int
    ) -> None:
        """Combine head."""
        if not len(self.combined_map[section_no][recommendation_no].keys()):
            self.combined_map[section_no][recommendation_no][tgt_col_profile] = ['profile']
            for col in range(1, sheet_helper.get_max_col() + 1):
                if col < tgt_col_profile:
                    tcol = col
                else:
                    tcol = col + 1
                value = sheet_helper.get_cell_value(row, col)
                self.combined_map[section_no][recommendation_no][tcol] = value

    def _combine_data(
        self, sheet_helper: SheetHelper, row: int, section_no: str, recommendation_no: str, tgt_col_profile: int
    ) -> None:
        """Combine data."""
        if not len(self.combined_map[section_no][recommendation_no].keys()):
            self.combined_map[section_no][recommendation_no][tgt_col_profile] = []
        sn = sheet_helper.get_sn()
        self.combined_map[section_no][recommendation_no][tgt_col_profile].append(sn)
        for col in range(1, sheet_helper.get_max_col() + 1):
            if col < tgt_col_profile:
                tcol = col
            else:
                tcol = col + 1
            self.combined_map[section_no][recommendation_no][tcol] = sheet_helper.get_cell_value(row, col)

    def _save(self) -> None:
        """Save."""
        self.wb.save(self.opath)
        logger.debug(f'{self.opath} saved')


class PropertyHelper:
    """Property Helper."""

    @staticmethod
    def normalize_name(name: str) -> str:
        """Normalize name."""
        rval = name.replace('(', '').replace(')', '').replace('#', '').strip()
        rval = rval.replace(' ', '_')
        rval = rval.replace('__', '_')
        return rval


class CsvHelper:
    """Csv Helper."""

    def __init__(self, path: pathlib.Path) -> None:
        """Initialize."""
        self.path = path
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.rows = []

    def add_row(self, row: List[str]) -> None:
        """Add row."""
        self.rows.append(row)

    def delete_last_row(self) -> None:
        """Delete last row."""
        self.rows = self.rows[:-1]

    def write(self) -> None:
        """Write csv file."""
        with open(self.path, 'w', newline='', encoding='utf-8') as output:
            csv_writer = csv.writer(output, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            for row in self.rows:
                csv_writer.writerow(row)

    def save(self) -> None:
        """Save csv file."""
        self.write()
        logger.debug(f'{self.path} saved')

    @staticmethod
    def columns() -> Dict:
        """Columns."""
        return {
            'Component_Title': 'A human readable name for the component.',  # noqa
            'Component_Description': 'A description of the component including information about its function.',  # noqa
            'Component_Type': 'A category describing the purpose of the component. ALLOWED VALUES interconnection:software:hardware:service:physical:process-procedure:plan:guidance:standard:validation:',  # noqa
            'Profile': 'List of CIS profiles',  # noqa
            'Rule_Id': 'A textual label that uniquely identifies a policy (desired state) that can be used to reference it elsewhere in this or other documents.',  # noqa
            'Rule_Description': 'A description of the policy (desired state) including information about its purpose and scope.',  # noqa
            'Profile_Source': 'A URL reference to the source catalog or profile for which this component is implementing controls for. A profile designates a selection and configuration of controls from one or more catalogs.',  # noqa
            'Profile_Description': 'A description of the profile.',  # noqa
            'Control_Id_List': 'A list of textual labels that uniquely identify the controls or statements that the component implements.',  # noqa
            'Namespace': 'A namespace qualifying the property\'s name. This allows different organizations to associate distinct semantics with the same name. Used in conjunction with "class" as the ontology concept.',  # noqa
        }


class CsvRowMgr:
    """Csv row manager."""

    def __init__(self, row_names: List) -> None:
        """Initialize."""
        self.row_names = row_names
        self.map = {}

    def put(self, key: str, val: str) -> None:
        """Put."""
        if key not in self.row_names:
            raise RuntimeError(f'{key} not found')
        self.map[key] = val

    def get(self) -> List[str]:
        """Get."""
        row = []
        for name in self.row_names:
            if name in self.map.keys():
                row.append(self.map[name])
            else:
                row.append('')
        return row


class CisControlsHelper:
    """Cis controls helper."""

    def __init__(self, cis_controls: str) -> None:
        """Initialize."""
        self.cis_controls = cis_controls

    def ctl_generator(self) -> Iterator[Dict]:
        """Generate ctls until finished."""
        parts = self.cis_controls.split(';TITLE:')
        # restore TITLE:
        for n in range(len(parts)):
            parts[n] = f'TITLE:{parts[n]}'
        # process triples TITLE, CONTROL, DESCRIPTION
        for part in parts:
            if part:
                s1 = part.split('DESCRIPTION:')
                description = s1[1].strip()
                s2 = s1[0].split('CONTROL:')
                control = s2[1].strip()
                control_version = control.split()[0]
                control_id = control.split()[1]
                s3 = s2[0].split('TITLE:')
                title = s3[1].strip()
                ctl = {
                    'description': description,
                    'control-version': control_version,
                    'control-id': control_id,
                    'title': title
                }
                yield ctl

    def get_ctl_list(self, ctl_pfx: str, profile_version: List[str]) -> List[str]:
        """Get control list."""
        ctl_list = []
        if self.cis_controls:
            for ctl in self.ctl_generator():
                if ctl['control-version'] not in profile_version:
                    continue
                ctl_id = ctl['control-id']
                try:
                    float(ctl_id)
                except Exception:
                    text = f'missing or invalid control-id: "{ctl_id}"'
                    raise RuntimeError(text)
                ctl_list.append(f'{ctl_pfx}{ctl_id}')
        return ctl_list


class XlsxToCsvHelper:
    """Xlsx to csv helper."""

    def __init__(self, config_object: SectionProxy, tmpdir: str) -> None:
        """Initialize."""
        self.config_object = config_object
        benchmark_file = self.config_object['benchmark-file']
        self.ipath = pathlib.Path(benchmark_file)
        self.xpath = pathlib.Path(tmpdir) / self.ipath.name
        path = pathlib.Path(tmpdir) / self.xpath.name
        self.opath = path.with_suffix('.csv')
        self.wb = load_workbook(self.xpath)
        # worksheet
        self.ws = self.wb[SheetHelper.get_sheetname()]
        self._create_maps()
        # excluded columns
        default_columns_exclude = [f'"{head_recommendation_no}"', '"Profile"', '"Description"']
        columns_exclude = self.config_object.get('columns-exclude')
        if columns_exclude:
            columns_exclude = columns_exclude.strip().split(',')
        else:
            columns_exclude = default_columns_exclude
        self._columns_exclude = []
        for col in columns_exclude:
            name = PropertyHelper.normalize_name(col.replace('"', ''))
            self._columns_exclude.append(name)
        # benchmark control prefix
        self._benchmark_control_prefix = self.config_object.get(
            'benchmark-control-prefix', default_benchmark_control_prefix
        )
        if not self._benchmark_control_prefix.endswith('-'):
            self._benchmark_control_prefix = f'{self._benchmark_control_prefix}-'
        # benchmark rule prefix
        self._benchmark_rule_prefix = self.config_object.get('benchmark-rule-prefix', default_benchmark_rule_prefix)
        if not self._benchmark_rule_prefix.endswith('-'):
            self._benchmark_rule_prefix = f'{self._benchmark_rule_prefix}-'

    def _create_maps(self) -> Dict:
        """Create maps."""
        self._map_col_key_to_number = {}
        self._map_name_to_col_key = {}
        row = 1
        cols = self.ws.max_column + 1
        for col in range(row, cols):
            cell = self.ws.cell(row, col)
            if cell.value:
                key = self._name_to_key(cell.value)
                self._map_col_key_to_number[key] = col
                self._map_name_to_col_key[cell.value] = key

    def _name_to_key(self, name: str) -> str:
        """Name to key."""
        rval = name.lower()
        return rval

    def _sanitize(self, value: str) -> str:
        """Sanitize value."""
        rval = value
        if value:
            rval = value.replace('\n', ' ')
        return rval

    def _get_map(self) -> Dict:
        """Get map."""
        return self._map_name_to_col_key

    def get_all_columns(self) -> List:
        """Get all columns."""
        return self._get_map().keys()

    def row_generator(self) -> Iterator[int]:
        """Generate rows until max reached."""
        row = 2
        while row <= self.ws.max_row:
            yield row
            row += 1

    def get(self, row: int, name: str) -> str:
        """Get cell value for given row and column name."""
        key = self._name_to_key(name)
        col = self._map_col_key_to_number[key]
        cell = self.ws.cell(row, col)
        return self._sanitize(cell.value)

    def is_same_rule(self, row_a: int, row_b: str) -> str:
        """Is same rule."""
        rule_a = self.get(row_a, head_recommendation_no)
        rule_b = self.get(row_b, head_recommendation_no)
        rval = rule_a == rule_b
        logger.debug(f'{rval} {rule_a} {rule_b}')
        return rval

    def is_excluded_column(self, column: str) -> bool:
        """Is excluded column."""
        for item in self._columns_exclude:
            if item.lower() == column.lower():
                return True
        return False

    def merge_row(self, prev_row: int, curr_row: int) -> bool:
        """Merge row."""
        rval = False
        if prev_row and self.is_same_rule(prev_row, curr_row):
            prof2 = self.get(prev_row, 'Profile')
            prof1 = self.get(curr_row, 'Profile')
            prof = f'"{prof1}; {prof2}"'
            self.csv_helper.delete_last_row()
            # col Profile
            self.csv_row_mgr.put('Profile', prof)
            # add body row
            row_body = self.csv_row_mgr.get()
            self.csv_helper.add_row(row_body)
            rval = True
        return rval

    def _is_column(self, c1: str, c2: str) -> bool:
        """Is column."""
        if c1.lower() == c2.lower():
            rval = True
        else:
            rval = False
        return rval

    def heading_row_1(self) -> None:
        """Heading row 1."""
        self.row_names = []
        for col_name in CsvHelper.columns().keys():
            name = PropertyHelper.normalize_name(col_name)
            self.row_names.append(name)

    def heading_row_2(self) -> None:
        """Heading row 2."""
        self.row_descs = []
        for col_desc in CsvHelper.columns().values():
            desc = col_desc
            self.row_descs.append(desc)
        # additional user columns
        for col in self.get_all_columns():
            name = PropertyHelper.normalize_name(col)
            if self.is_excluded_column(col):
                continue
            self.row_names.append(name)
            self.row_descs.append(name)
        # additional non-rule columns
        for col in self.non_rule_helper.get_all_columns():
            name = PropertyHelper.normalize_name(col)
            self.row_names.append(name)
            self.row_descs.append(name)

    def _get_ctl_list(self, prev_row: int, curr_row: int) -> List[str]:
        """Get_ctl_list."""
        ctl_list = []
        # if merged row, list is empty
        if not self.merge_row(prev_row, curr_row):
            # if non-rule row, list is empty
            rec_no = self.get(curr_row, head_recommendation_no)
            if rec_no is not None:
                # get list
                cis_controls = self.get(curr_row, 'CIS Controls')
                cis_control_helper = CisControlsHelper(cis_controls)
                ctl_list = cis_control_helper.get_ctl_list(
                    self._benchmark_control_prefix, self.config_object['profile-version']
                )
        return ctl_list

    def run(self) -> None:
        """Run."""
        self.csv_helper = CsvHelper(self.opath)
        self.non_rule_helper = NonRuleHelper(self.config_object, self)
        # heading row 1 - names
        self.heading_row_1()
        # heading row 2 - descriptions
        self.heading_row_2()
        # add heading rows
        self.csv_helper.add_row(self.row_names)
        self.csv_helper.add_row(self.row_descs)
        # body
        user_columns = []
        for col in self.get_all_columns():
            if self.is_excluded_column(col):
                continue
            user_columns.append(col)
        # process each data row of CIS Benchmark file
        prev_row = None
        for curr_row in self.row_generator():
            ctl_list = self._get_ctl_list(prev_row, curr_row)
            if not ctl_list:
                continue
            # create new row
            self.csv_row_mgr = CsvRowMgr(self.row_names)
            # col Component_Title
            self.csv_row_mgr.put('Component_Title', self.config_object['component-name'])
            # col Component_Description
            self.csv_row_mgr.put('Component_Description', self.config_object['component-description'])
            # col Component_Type
            self.csv_row_mgr.put('Component_Type', self.config_object['component-type'])
            # col Rule_ID
            rec_no = self.get(curr_row, head_recommendation_no)
            rule_id = f'{self._benchmark_rule_prefix}{rec_no}'
            self.csv_row_mgr.put('Rule_Id', rule_id)
            # col Rule_Description
            rule_desc = self.get(curr_row, 'Description')
            self.csv_row_mgr.put('Rule_Description', rule_desc)
            # col Profile_Source
            self.csv_row_mgr.put('Profile_Source', self.config_object['profile-source'])
            # col Profile_Description
            self.csv_row_mgr.put('Profile_Description', self.config_object['profile-description'])
            # col Control_Id_List
            ctl_list = ' '.join(ctl_list)
            self.csv_row_mgr.put('Control_Id_List', ctl_list)
            # col Namespace
            self.csv_row_mgr.put('Namespace', self.config_object['namespace'])
            # col Profile
            prof = self.get(curr_row, 'Profile')
            self.csv_row_mgr.put('Profile', prof)
            # col user
            for col in user_columns:
                val = self.get(curr_row, col)
                if val:
                    if self._is_column(col, 'cis controls'):
                        val = f'"{val}"'
                    name = PropertyHelper.normalize_name(col)
                    self.csv_row_mgr.put(name, val)
            # col group levels
            self.non_rule_helper.add_group_levels(rec_no, self.csv_row_mgr)
            # add body row
            row_body = self.csv_row_mgr.get()
            self.csv_helper.add_row(row_body)
            # previous row
            prev_row = curr_row
        # write body
        self.csv_helper.save()


class NonRuleHelper:
    """Non-rule Helper."""

    def __init__(self, config: SectionProxy, xlsx_helper: XlsxToCsvHelper) -> None:
        """Initialize."""
        self.col_prefix = 'Group'
        self.col_level = 'Level'
        self.config = config
        self.xlsx_helper = xlsx_helper
        default_columns_carry_forward = [head_section_no, 'Title', 'Description']
        self.columns_carry_forward = self.config.get('columns-carry-forward', default_columns_carry_forward)
        self.col_list = []
        self.sec_map = {}
        if self.columns_carry_forward:
            self._init_col_list()
            self._init_sec_map()

    def _init_col_list(self) -> None:
        """Init col list."""
        biggest = -1
        for row in self.xlsx_helper.row_generator():
            rec_no = self.xlsx_helper.get(row, head_recommendation_no)
            if rec_no:
                continue
            sec_no = self.xlsx_helper.get(row, head_section_no)
            count = sec_no.count('.')
            if count > biggest:
                biggest = count
        for i in range(biggest + 1):
            for col_name in self.columns_carry_forward:
                name = f'{self.col_prefix}_{col_name}_{self.col_level}_{i}'
                name = PropertyHelper.normalize_name(name)
                self.col_list.append(name)

    def _init_sec_map(self) -> None:
        """Init sec map."""
        for row in self.xlsx_helper.row_generator():
            rec_no = self.xlsx_helper.get(row, head_recommendation_no)
            if rec_no:
                continue
            sec_no = self.xlsx_helper.get(row, head_section_no)
            _map = {}
            for name in self.columns_carry_forward:
                _map[name] = self.xlsx_helper.get(row, name)
            self.sec_map[sec_no] = _map

    def get_all_columns(self) -> List:
        """Get all columns."""
        return self.col_list

    def add_group_levels(self, rec_no: str, csv_row_mgr: CsvRowMgr) -> None:
        """Add group levels."""
        parts = rec_no.split('.')[:-1]
        for i, part in enumerate(parts):
            if not i:
                sec_no = f'{part}'
            else:
                sec_no = f'{sec_no}.{part}'
            if sec_no in self.sec_map.keys():
                _map = self.sec_map[sec_no]
                for key in _map.keys():
                    val = _map[key]
                    name = f'{self.col_prefix}_{key}_{self.col_level}_{i}'
                    name = PropertyHelper.normalize_name(name)
                    csv_row_mgr.put(name, val)


class CsvToJsonHelper:
    """Csv to json helper."""

    def __init__(self, config_object: Optional[configparser.SectionProxy], tmpdir: str) -> None:
        """Initialize trestle task."""
        self.config_object = config_object
        benchmark_file = self.config_object['benchmark-file']
        self.ipath = pathlib.Path(benchmark_file)
        self.xpath = pathlib.Path(tmpdir) / self.ipath.name
        path = pathlib.Path(tmpdir) / self.xpath.name
        self.opath = path.with_suffix('.csv')
        self.config_object['csv-file'] = str(self.opath)
        self.csv_to_oscal_cd = CsvToOscalComponentDefinition(self.config_object)
        self.config_object['title'] = self.config_object['benchmark-title']
        self.config_object['version'] = self.config_object['benchmark-version']

    def run(self) -> None:
        """Run."""
        return self.csv_to_oscal_cd.execute()



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2024 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Testing mypy."""

import subprocess

from tests.test_utils import TEST_DIR

folder = TEST_DIR / 'trestle' / 'misc'
subject = folder / '_inventory.py'
config = folder / 'mypy.cfg'


def test_mypy(tmp_path) -> None:
    """Testing mypy."""
    cmd = ['trestle', 'version']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    output_bytes, _ = process.communicate()
    output = output_bytes.decode('utf-8')
    assert 'Trestle version' in output
    assert 'based on OSCAL version' in output
    #
    cmd = ['mypy', '--config-file', f'{config}', f'{subject}']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    output_bytes, _ = process.communicate()
    output = output_bytes.decode('utf-8')
    assert 'Success: no issues found in 1 source file' in output



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
#
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#                                                                        #
#        This file is automatically generated by gen_oscal.py            #
#                                                                        #
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#
#
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic.v1 import AnyUrl, EmailStr, Extra, Field, conint, constr, validator

from trestle.core.base_model import OscalBaseModel
from trestle.oscal import OSCAL_VERSION_REGEX, OSCAL_VERSION
import trestle.oscal.common as common
from trestle.oscal.common import RelatedObservation
from trestle.oscal.common import SystemComponent
from trestle.oscal.common import TaskValidValues
from trestle.oscal.common import TokenDatatype


class TermsAndConditions(OscalBaseModel):
    """
    Used to define various terms and conditions under which an assessment, described by the plan, can be performed. Each child part defines a different type of term or condition.
    """

    class Config:
        extra = Extra.forbid

    parts: Optional[List[common.AssessmentPart]] = Field(None)


class LocalDefinitions(OscalBaseModel):
    """
    Used to define data objects that are used in the assessment plan, that do not appear in the referenced SSP.
    """

    class Config:
        extra = Extra.forbid

    components: Optional[List[common.SystemComponent]] = Field(None)
    inventory_items: Optional[List[common.InventoryItem]] = Field(None, alias='inventory-items')
    users: Optional[List[common.SystemUser]] = Field(None)
    objectives_and_methods: Optional[List[common.LocalObjective]] = Field(None, alias='objectives-and-methods')
    activities: Optional[List[common.Activity]] = Field(None)
    remarks: Optional[str] = None


class AssessmentPlan(OscalBaseModel):
    """
    An assessment plan, such as those provided by a FedRAMP assessor.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this assessment plan in this or other OSCAL instances. The locally defined UUID of the assessment plan can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Assessment Plan Universally Unique Identifier',
    )
    metadata: common.Metadata
    import_ssp: common.ImportSsp = Field(..., alias='import-ssp')
    local_definitions: Optional[LocalDefinitions] = Field(
        None,
        alias='local-definitions',
        description=
        'Used to define data objects that are used in the assessment plan, that do not appear in the referenced SSP.',
        title='Local Definitions'
    )
    terms_and_conditions: Optional[TermsAndConditions] = Field(
        None,
        alias='terms-and-conditions',
        description=
        'Used to define various terms and conditions under which an assessment, described by the plan, can be performed. Each child part defines a different type of term or condition.',
        title='Assessment Plan Terms and Conditions'
    )
    reviewed_controls: common.ReviewedControls = Field(..., alias='reviewed-controls')
    assessment_subjects: Optional[List[common.AssessmentSubject]] = Field(None, alias='assessment-subjects')
    assessment_assets: Optional[common.AssessmentAssets] = Field(None, alias='assessment-assets')
    tasks: Optional[List[common.Task]] = Field(None)
    back_matter: Optional[common.BackMatter] = Field(None, alias='back-matter')


class Model(OscalBaseModel):
    assessment_plan: AssessmentPlan = Field(..., alias='assessment-plan')



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
#
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#                                                                        #
#        This file is automatically generated by gen_oscal.py            #
#                                                                        #
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#
#
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic.v1 import AnyUrl, EmailStr, Extra, Field, conint, constr, validator

from trestle.core.base_model import OscalBaseModel
from trestle.oscal import OSCAL_VERSION_REGEX, OSCAL_VERSION
import trestle.oscal.common as common


class Control(OscalBaseModel):
    """
    A structured object representing a requirement or guideline, which when implemented will reduce an aspect of risk related to an information system and its information.
    """

    class Config:
        extra = Extra.forbid

    id: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        description=
        'Identifies a control such that it can be referenced in the defining catalog and other OSCAL instances (e.g., profiles).',
        title='Control Identifier'
    )
    class_: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        alias='class',
        description='A textual label that provides a sub-type or characterization of the control.',
        title='Control Class'
    )
    title: str = Field(
        ...,
        description='A name given to the control, which may be used by a tool for display and navigation.',
        title='Control Title'
    )
    params: Optional[List[common.Parameter]] = Field(None)
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    parts: Optional[List[common.Part]] = Field(None)
    controls: Optional[List[Control]] = None


class Group(OscalBaseModel):
    """
    A group of controls, or of groups of controls.
    """

    class Config:
        extra = Extra.forbid

    id: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        description=
        'Identifies the group for the purpose of cross-linking within the defining instance or from other instances that reference the catalog.',
        title='Group Identifier'
    )
    class_: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        alias='class',
        description='A textual label that provides a sub-type or characterization of the group.',
        title='Group Class'
    )
    title: str = Field(
        ...,
        description='A name given to the group, which may be used by a tool for display and navigation.',
        title='Group Title'
    )
    params: Optional[List[common.Parameter]] = Field(None)
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    parts: Optional[List[common.Part]] = Field(None)
    groups: Optional[List[Group]] = None
    controls: Optional[List[Control]] = Field(None)


class Catalog(OscalBaseModel):
    """
    A structured, organized collection of control information.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
                 ) = Field(
                     ...,
                     description='Provides a globally unique means to identify a given catalog instance.',
                     title='Catalog Universally Unique Identifier'
                 )
    metadata: common.Metadata
    params: Optional[List[common.Parameter]] = Field(None)
    controls: Optional[List[Control]] = Field(None)
    groups: Optional[List[Group]] = Field(None)
    back_matter: Optional[common.BackMatter] = Field(None, alias='back-matter')


class Model(OscalBaseModel):
    catalog: Catalog



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for json and yaml manipulations of oscal files."""

import datetime
import pathlib

import pytest

from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError

from tests.test_utils import YAML_TEST_DATA_PATH

import trestle.common.const as const
import trestle.oscal.component as component

encoding = const.FILE_ENCODING


def test_yaml_load() -> None:
    """Test yaml load."""
    # happy path
    read_file = (YAML_TEST_DATA_PATH / 'good_simple.yaml').open('r', encoding=encoding)
    yaml = YAML(typ='safe')
    obj = yaml.load(read_file)
    assert obj is not None

    # unhappy path
    with pytest.raises(ParserError):
        read_file = (YAML_TEST_DATA_PATH / 'bad_simple.yaml').open('r', encoding=encoding)
        obj = yaml.load(read_file)


def test_yaml_dump(tmp_path: pathlib.Path) -> None:
    """Test yaml load and dump."""
    component_name = 'good_component.yaml'
    tmp_path = pathlib.Path(tmp_path)
    yaml = YAML(typ='safe')
    # happy path
    read_file = (YAML_TEST_DATA_PATH / component_name).open('r', encoding=encoding)
    component_obj = yaml.load(read_file)
    read_file.close()
    assert component_obj is not None

    dump_name = tmp_path / component_name
    write_file = dump_name.open('w', encoding=encoding)
    yaml.dump(component_obj, write_file)
    write_file.close()
    read_file = dump_name.open('r', encoding=encoding)
    saved_component = yaml.load(read_file)
    assert saved_component is not None

    assert saved_component == component_obj


def test_oscal_model(tmp_path: pathlib.Path) -> None:
    """Test pydantic oscal model."""
    good_component_name = 'good_component.yaml'
    tmp_path = pathlib.Path(tmp_path)

    tmp_path = pathlib.Path(tmp_path)

    # load good component
    read_file = YAML_TEST_DATA_PATH / good_component_name
    assert read_file.exists()
    component_obj = component.ComponentDefinition.oscal_read(read_file)
    assert component_obj is not None

    # write the oscal componet def out as yaml
    dump_name = tmp_path / good_component_name
    component_obj.oscal_write(dump_name)

    # read it back in
    component_reload = component.ComponentDefinition.oscal_read(dump_name)
    assert component_reload is not None

    # confirm same
    assert component_obj == component_reload

    # confirm it really is checking the time
    component_reload.metadata.last_modified = datetime.datetime.now()
    assert component_obj != component_reload

    # load good target with different timezone
    read_file = YAML_TEST_DATA_PATH / 'good_component_diff_tz.yaml'
    component_diff_tz = component.ComponentDefinition.oscal_read(read_file)
    assert component_diff_tz is not None

    # confirm same since different timezones but same utc time
    assert component_obj == component_diff_tz

    # try to load file with no timezone specified
    read_file = YAML_TEST_DATA_PATH / 'bad_component_no_tz.yaml'

    # confirm the load fails because it is invalid without timezone specified
    try:
        _ = component.ComponentDefinition.oscal_read(read_file)
    except Exception:
        assert True
    else:
        assert AssertionError()



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
#
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#                                                                        #
#        This file is automatically generated by gen_oscal.py            #
#                                                                        #
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#
#
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic.v1 import AnyUrl, EmailStr, Extra, Field, conint, constr, validator

from trestle.core.base_model import OscalBaseModel
from trestle.oscal import OSCAL_VERSION_REGEX, OSCAL_VERSION
import trestle.oscal.common as common
from trestle.oscal.common import RelatedObservation
from trestle.oscal.common import TaskValidValues
from trestle.oscal.common import TokenDatatype
from trestle.oscal.common import RelatedObservation as RelatedObservation1


class LocalDefinitions(OscalBaseModel):
    """
    Allows components, and inventory-items to be defined within the POA&M for circumstances where no OSCAL-based SSP exists, or is not delivered with the POA&M.
    """

    class Config:
        extra = Extra.forbid

    components: Optional[List[common.SystemComponent]] = Field(None)
    inventory_items: Optional[List[common.InventoryItem]] = Field(None, alias='inventory-items')
    assessment_assets: Optional[common.AssessmentAssets] = Field(None, alias='assessment-assets')
    remarks: Optional[str] = None


class Origination(OscalBaseModel):
    """
    Identifies the source of the finding, such as a tool or person.
    """

    class Config:
        extra = Extra.forbid

    actors: List[common.OriginActor] = Field(...)


class RelatedFinding(OscalBaseModel):
    """
    Relates the finding to referenced finding(s).
    """

    class Config:
        extra = Extra.forbid

    finding_uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        alias='finding-uuid',
        description='A machine-oriented identifier reference to a finding defined in the list of findings.',
        title='Finding Universally Unique Identifier Reference'
    )


class PoamItem(OscalBaseModel):
    """
    Describes an individual POA&M item.
    """

    class Config:
        extra = Extra.forbid

    uuid: Optional[constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    )] = Field(
        None,
        description=
        'A machine-oriented, globally unique identifier with instance scope that can be used to reference this POA&M item entry in this OSCAL instance. This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='POA&M Item Universally Unique Identifier'
    )
    title: str = Field(..., description='The title or name for this POA&M item .', title='POA&M Item Title')
    description: str = Field(
        ..., description='A human-readable description of POA&M item.', title='POA&M Item Description'
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    related_findings: Optional[List[RelatedFinding]] = Field(None, alias='related-findings')
    related_observations: Optional[List[common.RelatedObservation]] = Field(None, alias='related-observations')
    related_risks: Optional[List[common.RelatedRisk]] = Field(None, alias='related-risks')
    remarks: Optional[str] = None
    origins: Optional[List[Origination]] = Field(None)


class PlanOfActionAndMilestones(OscalBaseModel):
    """
    A plan of action and milestones which identifies initial and residual risks, deviations, and disposition, such as those required by FedRAMP.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with instancescope that can be used to reference this POA&M instance in this OSCAL instance. This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='POA&M Universally Unique Identifier'
    )
    metadata: common.Metadata
    import_ssp: Optional[common.ImportSsp] = Field(None, alias='import-ssp')
    system_id: Optional[common.SystemId] = Field(None, alias='system-id')
    local_definitions: Optional[LocalDefinitions] = Field(None, alias='local-definitions')
    observations: Optional[List[common.Observation]] = Field(None)
    risks: Optional[List[common.Risk]] = Field(None)
    findings: Optional[List[common.Finding]] = Field(None)
    poam_items: List[PoamItem] = Field(..., alias='poam-items')
    back_matter: Optional[common.BackMatter] = Field(None, alias='back-matter')


class Model(OscalBaseModel):
    plan_of_action_and_milestones: PlanOfActionAndMilestones = Field(..., alias='plan-of-action-and-milestones')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2024 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script to pre-process OSCAL schemas.

This module serves 3 purposes:

1. reordering of original OSCAL schema

The order of the "common" elements in the schemas produced by OSCAL is not consistent,
which leads to mismatches during Python code generation and less things appearing in common.
Reordering works nicely to increase the number of common objects.

2. reproducible names

Even with reordering, some names changed. Type1 in previous release of trestle is no longer
Type1 in current release. And the names were not necessarily meaningful. Code now navigates
to valid values in schema and assign names. Although some manual effort was initially
required, the anticipated OSCAL changes in the future are few. The end result is more
consistent naming.

3. allOf construct not correctly handled

The allOf construct, newly employed by the OSCAL schemas, is not handled correctly by the
Python code generator. An issue has been opened with the Python code generator.
See https://github.com/koxudaxi/datamodel-code-generator/issues/1901 (and others).
This part of pre-processing can be removed once a fix from the Python code generator
materializes.

"""

import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, List

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

schema_file_name_search_template = 'oscal_*_schema.json'

body_integrity_map = {}


def body_identical_check(token: str, title: str, content: Dict) -> None:
    """Check that common entries have identical bodies."""
    # derive common key
    key = token.split(':')[1] + title
    # if no common entry yet, add and return
    if key not in body_integrity_map.keys():
        body_integrity_map[key] = content
        logger.debug(f'initialize for {key}')
        return
    # check that existing common entry matches this entry
    entry = body_integrity_map[key]
    # existing and current content should be same size
    if len(content.keys()) != len(entry.keys()):
        logger.error(f'{token} size mismatch')
    # existing and current individual content should match
    for ekey in entry.keys():
        if ekey not in content.keys():
            logger.error(f'missing content key {token}')
        if entry[ekey] != content[ekey]:
            logger.error(f'mismatch content for key {token}')
        logger.debug(f'match for {token}')


def get_oscal_release(input_dir_name: str) -> str:
    """Get OSCAL release."""
    release = '?'
    input_dir_path = Path(input_dir_name)
    for full_name in input_dir_path.glob(schema_file_name_search_template):
        model_name = str(full_name)
        if 'catalog' in model_name:
            data = json_data_get(model_name)
            release = data['$id'].split('/')[-2]
    return release


def fixup_models(input_dir_name: str) -> Path:
    """Fix models."""
    # create fixup dir
    fixup_dir_name = f'tmp/{input_dir_name}-fixup'
    input_dir_path = Path(input_dir_name)
    fixup_dir_path = Path(fixup_dir_name)
    fixup_dir_path.mkdir(exist_ok=True, parents=True)
    # copy original schemas to fixup dir
    fixup_copy_schemas(input_dir_path, fixup_dir_path)
    # schema reorder entries
    fixup_json(fixup_dir_path)
    # schema patching
    patch_schemas(fixup_dir_path)
    patch_allof(fixup_dir_path)
    return fixup_dir_path


def fixup_copy_schemas(input_dir_path: Path, fixup_dir_path: Path) -> None:
    """Fixup copy schemas."""
    for full_name in input_dir_path.glob(schema_file_name_search_template):
        model_name = str(full_name)
        if 'complete' in model_name:
            continue
        cmd = f'cp -p {model_name} {str(fixup_dir_path)}'
        logger.debug(cmd)
        os.system(cmd)


def traverse_dict(data, model_name):
    """Recursively traverse dict and replace allOf with ref."""
    for _, value in data.items():
        if isinstance(value, dict):
            if 'allOf' in value.keys():
                allof = value['allOf'][1]
                if '$ref' in allof.keys():
                    value['$ref'] = allof['$ref']
                    del value['allOf']
            traverse_dict(value, model_name)


# - https://github.com/koxudaxi/datamodel-code-generator/issues/1901
def patch_allof(fixup_dir_path: Path) -> None:
    """Patch allOf in schemas."""
    for full_name in fixup_dir_path.glob(schema_file_name_search_template):
        model_name = str(full_name)
        data = json_data_get(model_name)
        traverse_dict(data, model_name)
        json_data_put(model_name, data)


# patch_schemas introduced for migrating from OSCAL 1.0.4 to 1.1.2 due to missing/broken
# support in datamodel-codegen tool. See issue(s):
# - https://github.com/koxudaxi/datamodel-code-generator/issues/1901
def patch_schemas(fixup_dir_path: Path) -> None:
    """Patch json schemas."""
    for full_name in fixup_dir_path.glob(schema_file_name_search_template):
        model_name = str(full_name)
        patch_finding_target(model_name)
        patch_poam_origins(model_name)
        patch_poam_item(model_name, 'related-findings')
        patch_poam_item(model_name, 'related-observations')
        patch_profile(model_name)
        create_refs(model_name)


def calculate_patch_key(key: str) -> str:
    """Calculate patch key."""
    patch_key = key
    patch_key = patch_key.replace('oscal-ap-', '')
    patch_key = patch_key.replace('oscal-ar-', '')
    patch_key = patch_key.replace('oscal-catalog-', '')
    patch_key = patch_key.replace('oscal-component-definition-', '')
    patch_key = patch_key.replace('oscal-poam-', '')
    patch_key = patch_key.replace('oscal-profile-', '')
    patch_key = patch_key.replace('oscal-ssp-', '')
    return patch_key


def patch_model(model_name: str, patch_json: Dict):
    """Patch model."""
    data = json_data_get(model_name)
    for key in data['definitions'].keys():
        patch_key = calculate_patch_key(key)
        if patch_key in patch_json.keys():
            logger.debug(f'patch: {model_name} {key}')
            data['definitions'][key] = patch_json[patch_key]
    json_data_put(model_name, data)


def patch_finding_target(model_name: str) -> None:
    """Patch finding target."""
    # finding target: change property name "status" to "objectiveStatus" to avoid conflict
    # later, oscal_normalize will do the reverse to get the field name correct
    data = json_data_get(model_name)
    for k1 in data['definitions'].keys():
        if not k1.endswith('-common:finding-target'):
            continue
        k2 = 'properties'
        if k2 not in data['definitions'][k1]:
            continue
        k3 = 'status'
        if k3 not in data['definitions'][k1][k2]:
            continue
        value = data['definitions'][k1][k2][k3]
        del data['definitions'][k1][k2][k3]
        u3 = 'objective_status'
        data['definitions'][k1][k2][u3] = value
        logger.debug(f'patch: {model_name} {k1}.{k2}.{k3} -> {k1}.{k2}.{u3}')
        json_data_put(model_name, data)


def patch_poam_origins(model_name: str) -> None:
    """Patch POAM origins."""
    # POAM: change property name "origins" to "originations" to avoid conflict
    # later, oscal_normalize will do the reverse to get the field name correct
    if not model_name.endswith('oscal_poam_schema.json'):
        return
    data = json_data_get(model_name)
    k1 = 'oscal-poam-oscal-poam:poam-item'
    if k1 not in data['definitions'].keys():
        return
    k2 = 'properties'
    if k2 not in data['definitions'][k1]:
        return
    k3 = 'origins'
    if k3 not in data['definitions'][k1][k2]:
        return
    value = data['definitions'][k1][k2][k3]
    del data['definitions'][k1][k2][k3]
    u3 = 'originations'
    data['definitions'][k1][k2][u3] = value
    logger.debug(f'patch: {model_name} {k1}.{k2}.{k3} -> {k1}.{k2}.{u3}')
    json_data_put(model_name, data)


# With this patch the description in POAM slightly altered to match the other models.
# This beneficially results in RelatedObservation being common in POAM (no need for RelatedObservation1).
def patch_poam_item(model_name: str, k3: str) -> None:
    """Patch POAM item."""
    # POAM: change related-finding item description from "poam-item" to "finding" to avoid conflict
    if not model_name.endswith('oscal_poam_schema.json'):
        return
    data = json_data_get(model_name)
    k1 = 'oscal-poam-oscal-poam:poam-item'
    if k1 not in data['definitions'].keys():
        return
    k2 = 'properties'
    if k2 not in data['definitions'][k1]:
        return
    if k3 not in data['definitions'][k1][k2]:
        return
    k4 = 'items'
    if k4 not in data['definitions'][k1][k2][k3]:
        return
    k5 = 'description'
    if k5 not in data['definitions'][k1][k2][k3][k4]:
        return
    old_value = 'poam-item'
    if old_value not in data['definitions'][k1][k2][k3][k4][k5]:
        return
    new_value = 'finding'
    data['definitions'][k1][k2][k3][k4][k5] = data['definitions'][k1][k2][k3][k4][k5].replace(old_value, new_value)
    logger.debug(f'patch: {model_name} {k1}.{k2}.{k3}.{k4}.{k5} {old_value} -> {new_value}')
    json_data_put(model_name, data)


# Profile: change name "select-control-by-id" to "select-control" to avoid conflict
# This does not cause in any field name changes.
def patch_profile(model_name: str) -> None:
    """Patch profile."""
    if not model_name.endswith('oscal_profile_schema.json'):
        return
    old_data = data_get(model_name)
    new_data = []
    old_value = 'select-control-by-id'
    new_value = 'select-control'
    count = 0
    for line in old_data:
        if old_value in line:
            count += 1
            line = line.replace(old_value, new_value)
        new_data.append(line)
    logger.debug(f'patch: {model_name} {old_value} -> {new_value}, count={count}')
    data_put(model_name, new_data)
    # format
    data = json_data_get(model_name)
    json_data_put(model_name, data)


def _find(needle: str, haystack: Dict) -> Any:
    """Find needle in haystack."""
    for item in haystack:
        if item == needle:
            return haystack[item]
    return None


# Create refs for common elements
# This beneficially results in unnamed entities getting assigned pre-determined names,
# rather than letting datamodel-codegen tool assign names nondeterministically.
def create_refs(model_name: str) -> None:
    """Create refs."""
    # Task Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:task',
        'oscal-ar-oscal-assessment-common:task',
        'oscal-poam-oscal-assessment-common:task',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'TaskValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Time Unit Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:task',
        'oscal-ar-oscal-assessment-common:task',
        'oscal-poam-oscal-assessment-common:task',
    ]
    navigation = ['properties', 'timing', 'properties', 'at-frequency', 'properties', 'unit', 'allOf']
    ref_name = 'TimeUnitValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Finding Target Type Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:finding-target',
        'oscal-ar-oscal-assessment-common:finding-target',
        'oscal-poam-oscal-assessment-common:finding-target',
    ]
    navigation = ['properties', 'type', 'allOf']
    ref_name = 'FindingTargetTypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Objective Status State Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:finding-target',
        'oscal-ar-oscal-assessment-common:finding-target',
        'oscal-poam-oscal-assessment-common:finding-target',
    ]
    navigation = ['properties', 'objective_status', 'properties', 'state', 'allOf']
    ref_name = 'ObjectiveStatusStateValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Threat Id Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:threat-id',
        'oscal-ar-oscal-assessment-common:threat-id',
        'oscal-poam-oscal-assessment-common:threat-id',
    ]
    navigation = ['properties', 'system', 'anyOf']
    ref_name = 'ThreatIdValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Select Subject By Id Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:select-subject-by-id',
        'oscal-ar-oscal-assessment-common:select-subject-by-id',
        'oscal-poam-oscal-assessment-common:select-subject-by-id',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'SelectSubjectByIdValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Assessment Subject Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:assessment-subject',
        'oscal-ar-oscal-assessment-common:assessment-subject',
        'oscal-poam-oscal-assessment-common:assessment-subject',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'AssessmentSubjectValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Naming System Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:characterization',
        'oscal-ar-oscal-assessment-common:characterization',
        'oscal-poam-oscal-assessment-common:characterization',
    ]
    navigation = ['properties', 'facets', 'items', 'properties', 'system', 'anyOf']
    ref_name = 'NamingSystemValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Subject Reference Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:subject-reference',
        'oscal-ar-oscal-assessment-common:subject-reference',
        'oscal-poam-oscal-assessment-common:subject-reference',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'SubjectReferenceValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Subject Reference Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:observation',
        'oscal-ar-oscal-assessment-common:observation',
        'oscal-poam-oscal-assessment-common:observation',
    ]
    navigation = ['properties', 'types', 'items', 'anyOf']
    ref_name = 'ObservationTypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Risk Status Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:risk-status',
        'oscal-ar-oscal-assessment-common:risk-status',
        'oscal-poam-oscal-assessment-common:risk-status',
    ]
    navigation = ['anyOf']
    ref_name = 'RiskStatusValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # How Many Valid Values
    list_ = [
        'oscal-ap-oscal-control-common:parameter-selection',
        'oscal-ar-oscal-control-common:parameter-selection',
        'oscal-catalog-oscal-control-common:parameter-selection',
        'oscal-component-definition-oscal-control-common:parameter-selection',
        'oscal-poam-oscal-control-common:parameter-selection',
        'oscal-profile-oscal-control-common:parameter-selection',
        'oscal-ssp-oscal-control-common:parameter-selection',
    ]
    navigation = ['properties', 'how-many', 'allOf']
    ref_name = 'HowManyValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Telephone Type Valid Values
    list_ = [
        'oscal-ap-oscal-metadata:telephone-number',
        'oscal-ar-oscal-metadata:telephone-number',
        'oscal-catalog-oscal-metadata:telephone-number',
        'oscal-component-definition-oscal-metadata:telephone-number',
        'oscal-poam-oscal-metadata:telephone-number',
        'oscal-profile-oscal-metadata:telephone-number',
        'oscal-ssp-oscal-metadata:telephone-number',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'TelephoneTypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Address Type Valid Values
    list_ = [
        'oscal-ap-oscal-metadata:address',
        'oscal-ar-oscal-metadata:address',
        'oscal-catalog-oscal-metadata:address',
        'oscal-component-definition-oscal-metadata:address',
        'oscal-poam-oscal-metadata:address',
        'oscal-profile-oscal-metadata:address',
        'oscal-ssp-oscal-metadata:address',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'AddressTypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # External Scheme Valid Values
    list_ = [
        'oscal-ap-oscal-metadata:metadata',
        'oscal-ar-oscal-metadata:metadata',
        'oscal-catalog-oscal-metadata:metadata',
        'oscal-component-definition-oscal-metadata:metadata',
        'oscal-poam-oscal-metadata:metadata',
        'oscal-profile-oscal-metadata:metadata',
        'oscal-ssp-oscal-metadata:metadata',
    ]
    navigation = [
        'properties', 'parties', 'items', 'properties', 'external-ids', 'items', 'properties', 'scheme', 'anyOf'
    ]
    ref_name = 'ExternalSchemeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Party Type Valid Values
    list_ = [
        'oscal-ap-oscal-metadata:metadata',
        'oscal-ar-oscal-metadata:metadata',
        'oscal-catalog-oscal-metadata:metadata',
        'oscal-component-definition-oscal-metadata:metadata',
        'oscal-poam-oscal-metadata:metadata',
        'oscal-profile-oscal-metadata:metadata',
        'oscal-ssp-oscal-metadata:metadata',
    ]
    navigation = ['properties', 'parties', 'items', 'properties', 'type', 'allOf']
    ref_name = 'Party TypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Document Scheme Valid Values
    list_ = [
        'oscal-ap-oscal-metadata:document-id',
        'oscal-ar-oscal-metadata:document-id',
        'oscal-catalog-oscal-metadata:document-id',
        'oscal-component-definition-oscal-metadata:document-id',
        'oscal-poam-oscal-metadata:document-id',
        'oscal-profile-oscal-metadata:document-id',
        'oscal-ssp-oscal-metadata:document-id',
    ]
    navigation = ['properties', 'scheme', 'anyOf']
    ref_name = 'DocumentSchemeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Defined Component Type Valid Values
    list_ = [
        'oscal-component-definition-oscal-component-definition:defined-component',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'DefinedComponentTypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # System Component Type Valid Values
    list_ = [
        'oscal-ap-oscal-implementation-common:system-component',
        'oscal-ar-oscal-implementation-common:system-component',
        'oscal-catalog-oscal-implementation-common:system-component',
        'oscal-component-definition-oscal-implementation-common:system-component',
        'oscal-poam-oscal-implementation-common:system-component',
        'oscal-profile-oscal-implementation-common:system-component',
        'oscal-ssp-oscal-implementation-common:system-component',
    ]
    navigation = ['properties', 'type', 'anyOf']
    ref_name = 'SystemComponentTypeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # System Component Operational State
    list_ = [
        'oscal-ap-oscal-implementation-common:system-component',
        'oscal-ar-oscal-implementation-common:system-component',
        'oscal-catalog-oscal-implementation-common:system-component',
        'oscal-component-definition-oscal-implementation-common:system-component',
        'oscal-poam-oscal-implementation-common:system-component',
        'oscal-profile-oscal-implementation-common:system-component',
        'oscal-ssp-oscal-implementation-common:system-component',
    ]
    navigation = ['properties', 'status', 'properties', 'state', 'allOf']
    ref_name = 'SystemComponentOperationalStateValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Origin Actor Type Valid Values
    list_ = [
        'oscal-ap-oscal-assessment-common:origin-actor',
        'oscal-ar-oscal-assessment-common:origin-actor',
        'oscal-poam-oscal-assessment-common:origin-actor',
    ]
    navigation = ['properties', 'type', 'allOf']
    ref_name = 'OriginActorValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Port Range Valid Values
    list_ = [
        'oscal-ap-oscal-implementation-common:port-range',
        'oscal-ar-oscal-implementation-common:port-range',
        'oscal-component-definition-oscal-implementation-common:port-range',
        'oscal-poam-oscal-implementation-common:port-range',
        'oscal-ssp-oscal-implementation-common:port-range',
    ]
    navigation = ['properties', 'transport', 'allOf']
    ref_name = 'PortRangeValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Operational State Valid Values
    list_ = [
        'oscal-ssp-oscal-ssp:status',
    ]
    navigation = ['properties', 'state', 'allOf']
    ref_name = 'OperationalStateValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Item Name Valid Values
    list_ = [
        'oscal-profile-oscal-profile:modify',
    ]
    navigation = [
        'properties', 'alters', 'items', 'properties', 'removes', 'items', 'properties', 'by-item-name', 'allOf'
    ]
    ref_name = 'ItemNameValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Order Valid Values
    list_ = [
        'oscal-profile-oscal-profile:insert-controls',
    ]
    navigation = ['properties', 'order', 'allOf']
    ref_name = 'OrderValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # With Child Controls Valid Values
    list_ = [
        'oscal-profile-oscal-profile:select-control',
    ]
    navigation = ['properties', 'with-child-controls', 'allOf']
    ref_name = 'WithChildControlsValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Position Valid Values
    list_ = [
        'oscal-profile-oscal-profile:modify',
    ]
    navigation = ['properties', 'alters', 'items', 'properties', 'adds', 'items', 'properties', 'position', 'allOf']
    ref_name = 'PositionValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)
    # Combination Method Valid Values
    list_ = [
        'oscal-profile-oscal-profile:merge',
    ]
    navigation = ['properties', 'combine', 'properties', 'method', 'allOf']
    ref_name = 'CombinationMethodValidValues'
    for root in list_:
        create_ref(model_name, root, navigation, ref_name)


def _fetch(tgt: Any, key: str) -> Any:
    """Fetch."""
    try:
        return tgt.get(key)
    except Exception:
        return _find(key, tgt)


def _get_title(model_name: str, root: str, navigation: List[str]) -> str:
    """Get title."""
    title = None
    data = json_data_get(model_name)
    tgt = data['definitions']
    tgt = tgt.get(root)
    if not tgt:
        return
    for leaf in navigation:
        tgt = _fetch(tgt, leaf)
    title = tgt
    return title


def create_ref(model_name: str, root: str, navigation: List[str], ref_name: str) -> None:
    """Create ref."""
    data = json_data_get(model_name)
    tgt = data['definitions']
    tgt = tgt.get(root)
    if not tgt:
        return
    for leaf in navigation:
        tgt = _fetch(tgt, leaf)
    item = tgt[1]
    replacement = {'$ref': f'#/definitions/{ref_name}'}
    tgt[1] = replacement
    tgt = data['definitions']
    tgt[ref_name] = item
    logger.debug(f'patch: {model_name} {replacement}')
    # title
    title_navigation = navigation
    title_navigation[-1] = 'title'
    title = _get_title(model_name, root, navigation)
    body_identical_check(root, title, replacement)
    json_data_put(model_name, data)


def data_get(model_name: str) -> List:
    """Get data."""
    data = []
    with open(model_name, 'r') as f:
        for line in f:
            data.append(line.strip())
    return data


def data_put(model_name: str, data: List) -> None:
    """Put data."""
    with open(model_name, 'w') as f:
        for line in data:
            f.write(f'{line}\n')


def json_data_get(model_name: str) -> Dict:
    """Get json data."""
    with open(model_name, 'r') as f:
        data = json.load(f)
    return data


def json_data_put(model_name: str, data: Dict) -> None:
    """Put json data."""
    with open(model_name, 'w') as f:
        json_object = json.dumps(data, indent=2)
        f.write(json_object)


def fixup_json(fixup_dir_path: Path) -> None:
    """Fixup json."""
    for full_name in fixup_dir_path.glob(schema_file_name_search_template):
        model_name = str(full_name)
        data = json_data_get(model_name)
        names_reorder(data)
        json_data_put(model_name, data)


def get_order(key: str) -> int:
    """Get order."""
    if key != 'range':
        if 'json-schema-directive' in key:
            return 0
        if key.endswith('ap:assessment-plan'):
            return 1
        if key.endswith('ar:assessment-results'):
            return 1
        if key.endswith('catalog:catalog'):
            return 1
        if key.endswith('component-definition:component-definition'):
            return 1
        if key.endswith('poam:plan-of-action-and-milestones'):
            return 1
        if key.endswith('profile:profile'):
            return 1
        if key.endswith('ssp:system-security-plan'):
            return 1
        if 'oscal-metadata:' in key:
            return 2
        if 'oscal-control-common:' in key:
            return 3
        if 'oscal-implementation-common:' in key:
            return 4
        if 'oscal-assessment-common:' in key:
            return 5
        return 6
    else:
        # range = 1 + the highest number above
        return 7


def names_reorder(data: Dict) -> None:
    """Reorder."""
    reorder_defs = {}
    for index in range(get_order('range')):
        for key in data['definitions'].keys():
            order = get_order(key)
            if order == index:
                reorder_defs[key] = data['definitions'][key]
    len_old = len(data['definitions'])
    len_new = len(reorder_defs)
    if len_old != len_new:
        raise RuntimeError(f'old: {len_old} new: {len_new}')
    data['definitions'] = reorder_defs
